import pickle
import functools
import abc
import warnings

from pprint import pprint
from pathlib import Path
from tqdm import tqdm
#from tqdm.contrib.concurrent import process_map
from multiprocessing import Pool

import numpy as np
import pandas as pd
from pandas.core.common import SettingWithCopyWarning
warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

import dolphindb as ddb
import dolphindb.settings as keys

import sqlalchemy as sa
from .DDBLoader import DDBLoader


class DDBDailyLoader(DDBLoader):

    ddb_path = "dfs://daily_stock_ts"
    ddb_dbname = "db_daily_stock_ts"

    daily_kline_cols = [
        'code', 'm_nDate', 
        'open', 'high', 'low', 'close', 'vol',
        'amount', 'cjbs', 'yclose', 
        'PctChg', 'IsZt', 'IsDt', 'IsST', 'IsGoDelist',
        'FloatShares', 'MarketValues',
        'factor'
    ]

    daily_kline_col_types = [
        'SYMBOL', 'DATE',
        'DOUBLE', 'DOUBLE', 'DOUBLE', 'DOUBLE', 'DOUBLE', 
        'DOUBLE', 'INT', 'DOUBLE',
        'DOUBLE', 'INT', 'INT', 'INT', 'INT',
        'DOUBLE', 'DOUBLE',
        'DOUBLE'
    ]

    memory_table_name = 'daily_kline_mem'
    partition_table_name = 'daily_kline'

    def create_ddb_database(self):
        # TODO: daily数据库已经在DDBDailyFactor中被创建了
        #       后续可以迁移过来，不过现在就暂时先不管了
        pass


    def load_ddb_database(self):
        self.ddb_sess.run("""
            {dbName} = database(directory='{dbPath}')
        """.format(
            dbName = self.ddb_dbname,
            dbPath = self.ddb_path
        ))
        print('Did load database from', self.ddb_path)


    def create_ddb_partition_table(self):
        # TODO: 现在只做一个日频行情数据表，今后可能考虑把基本面数据也迁移过来

        # 由于日频行情数据的表结构相对简单，所以直接把表结构写在这里代码里即可
        # 搬迁数据的时候需要考虑按照逐个股票来搬迁，以免造成对内存的巨大压力
        self.ddb_sess.run("""
            // 确保删除原表
            if (existsTable("{ddb_daily_path}", "{partition_table_name}")) {{
                dropTable({ddb_daily_dbname}, "{partition_table_name}");
            }}

            // 然后根据内存表的结构，创建持久化的分区表
            {partition_table_name} = {ddb_daily_dbname}.createPartitionedTable(
                table = {memory_table_name},
                tableName = "{partition_table_name}",
                partitionColumns = `code,
                sortColumns = `code`m_nDate,
                compressMethods = {{m_nDate:"delta"}}
            );
        """.format(
            ddb_daily_path = self.ddb_path,
            ddb_daily_dbname = self.ddb_dbname,
            memory_table_name = self.memory_table_name,
            partition_table_name = self.partition_table_name,
        ))
    

    def create_ddb_memory_table(self, capacity):
        self.ddb_sess.run("""
            // 先创建一个空的内存表用来表征结构，如果无需插入数据，capacity可以设为10
            {memory_table_name} = table({capacity}:0, {col_names}, [{col_types}]);
        """.format(
            memory_table_name = self.memory_table_name,
            capacity = capacity,
            col_names =  '`' + '`'.join(self.daily_kline_cols),
            col_types = ', '.join(self.daily_kline_col_types)
        ))


    def dump_to_ddb(self):
        # 先创建一个分区表，然后再逐个股票的数据插入
        # 1. 需要额外控制在插入第一个股票数据的时候创建分区表比较麻烦
        # 2. python程序中的dataframe直接上传到dolphindb内存表，不需要考虑内存表字段类型，分区表中设置好即可


        self.create_ddb_memory_table(10)
        print('Did create ddb memory table.')
        pprint(self.ddb_sess.run(f"schema({self.memory_table_name})"))
        self.create_ddb_partition_table()
        print('Did create ddb partition table.')
        pprint(self.ddb_sess.run(f"schema({self.partition_table_name})"))

        with self.mssql_engine.connect() as conn:
            stat = "select distinct [StockID] from [StockDaily].dbo.[DailyKLine]"
            stock_id_list = list(conn.execute(stat).fetchall())

            with tqdm(stock_id_list) as pbar:
                for (stock_id,) in pbar:
                    pbar.set_description(f"Will work on {stock_id}") 
                    #pbar.set_description(f"Will fetch all data of {stock_id} from SQL Server")
                    stat = """
                        select * from [StockDaily].dbo.[DailyKLine]
                        where StockID='{stock_id}'
                    """.format(
                        stock_id = stock_id
                    )
                    row_list = list(conn.execute(stat).fetchall())
                    num_rows = len(row_list)

                    #pbar.set_description(f"Will work on dumping job on {stock_id} of len {num_rows}")
                    df = pd.DataFrame(row_list)
                    df['date'] = DDBLoader.make_date(df['date'])
                    df['StockID'] = DDBLoader.tscode_to_windcode(df['StockID'])
                    self.ddb_sess.upload({self.memory_table_name : df})
                    #print('Did upload dataframe to ddb.')
                    #pprint(self.ddb_sess.run(f"schema({self.memory_table_name})"))
                    #break
                    self.ddb_sess.run(f"{self.partition_table_name}.tableInsert({self.memory_table_name})")


class DDBDailyFactorLoader(DDBDailyLoader):

    daily_kline_cols = [
        'code', 'm_nDate',
        # 4种量价配合的因子
        'trend_with_turnover', 'trend_with_amount',
        "abs_trend_with_turnover", "abs_trend_with_amount",
        # Alpha101中，量价背离的因子
        "alpha101_22"
    ]

    daily_kline_col_types = [
        'SYMBOL', 'DATE',
        'DOUBLE', 'DOUBLE',
        'DOUBLE', 'DOUBLE',
        'DOUBLE'
    ]


    memory_table_name = 'daily_factor_mem'
    partition_table_name = 'daily_factor'


    def dump_to_ddb(self):
        self.create_ddb_memory_table(10)
        print('Did create ddb memory table.')
        pprint(self.ddb_sess.run(f"schema({self.memory_table_name})"))

        self.create_ddb_partition_table()
        print('Did create ddb partition table.')
        pprint(self.ddb_sess.run(f"schema({self.partition_table_name})"))

        df_list = [
            self.alpha101_22,
            self.trend_with_amount,
        ]
        df = pd.concat(df_list, axis=1)
        df.reset_index(inplace=True)
        print('Did prepare the dataframe for insertion:')
        print(df.head())

        self.ddb_sess.upload({"tbl": df})
        self.ddb_sess.run("tableInsert(loadTable('dfs://daily_stock_ts', 'daily_factor'), tbl)")
        pprint("Did dump data to partition table.")


    @property
    def alpha101_22(self):

        sql = """
            vol_20 = select code, m_nDate, mstdp(close, 20, 9) as vol_20
                from loadTable("dfs://daily_stock_ts", "daily_kline")
                context by code;

            rank_vol_20 = select code, m_nDate, rank(vol_20, tiesMethod='average', percent=true)
                from vol_20 context by m_nDate;

            corr_5 = select code, m_nDate, mcorr(high, vol, 5) as corr_5
                from loadTable("dfs://daily_stock_ts", "daily_kline")
                context by code;

            delta_corr_5 = select code, m_nDate, mfirst(corr_5, 5) - corr_5 as delta_corr_5
                from corr_5 context by code;

            alpha101_22 = select code, m_nDate, delta_corr_5 * rank_vol_20 as alpha101_22
                from ej(rank_vol_20, delta_corr_5, `code`m_nDate);

            alpha101_22;
            """

        df = self.ddb_sess.run(sql)
        df.set_index(["code", "m_nDate"], inplace=True)
        return df


    @property
    def trend_with_amount(self):

        factor_list = [
            'trend_with_turnover', 'trend_with_amount',
            "abs_trend_with_turnover", "abs_trend_with_amount",
        ]

        factor_def = {
            "trend_with_turnover": "(rank(amount/MarketValues, tiesMethod='average', percent=true) - 0.5) * winsorize(PctChg/20, 0.05)",
            "trend_with_amount": "(rank(amount, tiesMethod='average', percent=true) - 0.5) * winsorize(PctChg/20, 0.05)",
            "abs_trend_with_turnover": "(rank(amount/MarketValues, tiesMethod='average', percent=true) - 0.5) * abs(winsorize(PctChg/20, 0.05))",
            "abs_trend_with_amount": "(rank(amount, tiesMethod='average', percent=true) - 0.5) * abs(winsorize(PctChg/20, 0.05))",
        }

        cols = ", ".join([
            f"{factor_def[factor_name]} as {factor_name}" \
                    for factor_name in factor_list
        ])

        sql = f"""
            select code, m_nDate, {cols} 
                from loadTable("dfs://daily_stock_ts", "daily_kline")
                context by m_nDate
        """
        print("factor sql for trend with amount: " + sql)

        df = self.ddb_sess.run(sql)
        df.set_index(["code", "m_nDate"], inplace=True)
        return df


def main():

    # TODO: 
    # 可以使用`Fire`库，对函数调用再做一次封装，就可以避免每次运行不同参数时候需要修改内部多处的代码。

    # 日频行情数据
    #loader = DDBDailyLoader()

    loader = DDBDailyFactorLoader()
    loader.load_ddb_database()
    loader.dump_to_ddb()


if __name__ == '__main__':
    main() 

