import pickle
import functools
import warnings

from pprint import pprint
from pathlib import Path
from tqdm import tqdm
from multiprocessing import Pool

import numpy as np
import pandas as pd

import dolphindb as ddb
import dolphindb.settings as keys

import sqlalchemy as sa

from .DDBLoader import DDBLoader


class DDBIndexLoader(DDBLoader):

    ddb_path = "dfs://daily_stock_ts"
    ddb_dbname = "db_daily_stock_ts"


    def __init__(self, dtype, **kwargs):
        # TODO: 后续版本中，父类的构造函数里可能会增加一些设置项
        super().__init__(**kwargs)
        self.dtype = dtype

        if dtype == "concept":
            self.mem_tbl_name = "mem_idx_daily_concept"
            self.part_tbl_name ="idx_daily_concept"
        elif dtype == "kline":
            self.mem_tbl_name = "mem_idx_daily_kline"
            self.part_tbl_name ="idx_daily_kline"
        else:
            raise NotImplementedError(f"Unsupported `dtype` argument: {dtype}")

        self.make_fields()
        self.make_calendar_df()


    def make_fields(self):
        if self.dtype == "concept":
            with self.mssql_engine.connect() as conn:
                rs = conn.execute("select IndexID from [IndexInfo].[dbo].[Constituents] group by IndexID")
                self.fields = [index_id for (index_id,) in rs.fetchall()]
        elif self.dtype == "kline":
            self.fields = ['open', 'high', 'low', 'close', 'vol', 'amount', 'yclose']


    def make_calendar_df(self):
        # 这里我们使用天软日K先数据表来构造交易日历
        with self.mssql_engine.connect() as conn:
            if self.dtype == "concept":
                stat = "select [StockID], [date] from [StockDaily].[dbo].[DailyKLine] group by [StockID], [date]"
            elif self.dtype == "kline":
                stat = "select [StockID], [date] from [IndexDaily].[dbo].[DailyKLine] group by [StockID], [date]"
            else:
                raise NotImplementedError(f"Unsupported dtype: {self.dtype}")

            rs = conn.execute(stat)
            stock_date_list = [(stock_name, date) for stock_name, date in rs.fetchall()]

        self.df_calendar = pd.DataFrame(stock_date_list, columns=['code', 'm_nDate'])
        self.df_calendar['m_nDate'] = self.make_date(self.df_calendar['m_nDate'])
        self.df_calendar['code'] = self.tscode_to_windcode(self.df_calendar['code'])

        print('Did make the DataFrame for calendar')
        print(self.df_calendar.head())


    def load_ddb_database(self):
        self.ddb_sess.run("""
            {dbName} = database(directory='{dbPath}')
        """.format(
            dbName = self.ddb_dbname,
            dbPath = self.ddb_path
        ))
        print('Did load database from', self.ddb_path)


    def create_ddb_partition_table(self):
        if self.dtype == "concept":
            self._create_ddb_memory_table_concept()
        elif self.dtype == "kline":
            self._create_ddb_memory_table_kline()

        pprint(f"Did create memory table: {self.mem_tbl_name}")
        #res = self.ddb_sess.run(f"schema({mem_tbl_name}).colDefs")

        if self.ddb_sess.existsTable(self.ddb_path, self.part_tbl_name):
            pprint(f"Will drop partition table: {self.part_tbl_name}")
            self.ddb_sess.dropTable(self.ddb_path, self.part_tbl_name)

        self.ddb_sess.run("""
            {part_tbl_name} = {ddb_dbname}.createPartitionedTable(
                table = {mem_tbl_name},
                tableName = `{part_tbl_name},
                partitionColumns = `code,
                sortColumns = `code`m_nDate,
                compressMethods = {{m_nDate:"delta"}}
            )
        """.format(
            ddb_dbname      = self.ddb_dbname,
            part_tbl_name   = self.part_tbl_name,
            mem_tbl_name    = self.mem_tbl_name
        ))


    def _create_ddb_memory_table_concept(self):

        concept_list = self.fields
        col_name_list = ['code', 'm_nDate'] + concept_list
        col_type_list = ['SYMBOL', 'DATE'] + ['BOOL'] * len(concept_list)
        code = """
            {mem_tbl_name} = table(
                {capacity}:0, 
                {col_names}, 
                [{col_types}]
            );
        """.format(
            mem_tbl_name    =   self.mem_tbl_name,
            capacity        =   10,
            col_names       =   '`' + '`'.join(col_name_list),
            col_types       =   ','.join(col_type_list)
        )
        pprint(f"Will create mem table by:\n{code}")
        self.ddb_sess.run(code)


    def _create_ddb_memory_table_kline(self):

        col_name_list = ['code', 'm_nDate'] + self.fields
        col_type_list = ['SYMBOL', 'DATE'] + ['DOUBLE'] * len(self.fields)
        code = """
            {mem_tbl_name} = table(
                {capacity} : 0,
                {col_names},
                [{col_types}]
            );
        """.format(
            mem_tbl_name = self.mem_tbl_name,
            capacity = 10,
            col_names = '`' + '`'.join(col_name_list),
            col_types = ','.join(col_type_list)
        )
        pprint(f"Will create mem table by:\n{code}")
        self.ddb_sess.run(code)


    def _make_idx_daily_kline(self):
        with tqdm(self.df_calendar.groupby('code')) as pbar:
            for wind_code, df_calendar_stock in pbar:
                pbar.set_description(f"Will work on {wind_code}")

                 # 生成ts-code，用于查询Sql-Server中天软的概念板块指数
                ts_code = wind_code[-2:] + wind_code[:-3]

                df_calendar_stock.set_index(['code', 'm_nDate'], inplace=True)

                with self.mssql_engine.connect() as conn:
                    code = """
                        select 
                            {field_list}
                        from 
                            [IndexDaily].[dbo].[DailyKLine]
                        where
                            StockID='{index_id}'
                    """.format(
                        field_list = ','.join([f"[{field}]" for field in (['StockID', 'date'] + self.fields)]),
                        index_id = ts_code
                    )
                    rs = conn.execute(code)
                    row_list = rs.fetchall()

                df = pd.DataFrame(row_list, columns=['code', 'm_nDate'] + self.fields)
                df['code'] = self.tscode_to_windcode(df['code'])
                df['m_nDate'] = self.make_date(df['m_nDate'])
                df.set_index(['code', 'm_nDate'], inplace=True)

                yield wind_code, df


    def dump_idx_daily_kline_to_ddb(self):

        for idx_id, df in self._make_idx_daily_kline():
            df.reset_index(inplace=True)
            #pprint(f"Will append to partiton table: \n{df}")
            self.ddb_sess.upload({self.mem_tbl_name : df})
            self.ddb_sess.run("""
                append!(loadTable('{dbPath}', `{part_tbl_name}), {mem_tbl_name})
            """.format(
                dbPath          = self.ddb_path,
                part_tbl_name   = self.part_tbl_name,
                mem_tbl_name    = self.mem_tbl_name
            ))


    @staticmethod
    def _mark_stock2concept_onehot(df_stock2concept, concept_id, start_date, end_date):
        # 个股成为某个概念（指数）的起始日期是必定会提供的
        #   但是截止日期可能缺失，确实一般意味着当前仍然是在此概念板块中
        #   因此会通过将日期填充至最后一日来表示当前仍然在此概念板块内
        if end_date is None or end_date == 0:
            start_date = pd.to_datetime(str(start_date), format='%Y%m%d')
            df_stock2concept.loc[df_stock2concept.index.get_level_values('m_nDate') >= start_date] = True
        else:
            start_date = pd.to_datetime(str(start_date), format='%Y%m%d')
            end_date = pd.to_datetime(str(end_date), format='%Y%m%d')
            df_stock2concept.loc[
                (df_stock2concept.index.get_level_values('m_nDate') >= start_date) & 
                (df_stock2concept.index.get_level_values('m_nDate') <= end_date)
            ] = True


    def _make_stock2concept_onehot(self):

        # 从calendar中截取出与当前stock有关的日期，然后设置成index
        # 此处calendar使用的是海通高频数据构建，因此股票代码为WIND-CODE

        # 对calendar根据股票代码进行分组
        with tqdm(self.df_calendar.groupby('code')) as pbar:
            for wind_code, df_calendar_stock in pbar:
                pbar.set_description(f"Will work on {wind_code}")

                # 生成ts-code，用于查询Sql-Server中天软的概念板块指数
                ts_code = wind_code[-2:] + wind_code[:-3]

                df_calendar_stock.set_index(['code', 'm_nDate'], inplace=True)

                # 纵表转横表，`concept_list`作为列名
                df_stock2concept = pd.DataFrame(
                    False,  # one-hot横表，初始化都是0，后续根据Sql-Server的进出日期标注1
                    index=df_calendar_stock.index, 
                    columns=concept_list,
                    dtype="bool"
                )

                # 从Sql-Server中读取`stock_id`所对应的概念板块进出日期
                # 此数据是从天软指数数据中提取，因此需要使用TSCODE
                with self.mssql_engine.connect() as conn:
                    code = """
                        select 
                            SecId, IndexID, EnterDate, ExitDate 
                        from 
                            [IndexInfo].[dbo].[Constituents] 
                        where 
                            SecID='{stock_id}'
                    """.format(
                        stock_id = ts_code
                    )
                    rs = conn.execute(code)
                    row_list = rs.fetchall()

                # 从Sql-Server读取出单个股票的所有板块进出日期后，开始进行标记
                for (stock_id, concept_id, start_date, end_date) in row_list:
                    # mark the one-hot position one-by-one
                    self._mark_stock2concept_onehot(
                        df_stock2concept[concept_id], 
                        concept_id, 
                        start_date, end_date
                    )

                # yield the marked one-hot dataframe for one stock
                yield wind_code, df_stock2concept


    def dump_idx_concept_to_ddb(self):

        concept = self.fields

        for stock_id, df in self._make_stock2concept_onehot(concept_list):
            df.reset_index(inplace=True)
            self.ddb_sess.upload({self.mem_tbl_name : df})
            self.ddb_sess.run("""
                append!(loadTable('{dbPath}', `{part_tbl_name}), {mem_tbl_name})
            """.format(
                dbPath          = self.ddb_path,
                part_tbl_name   = self.part_tbl_name,
                mem_tbl_name    = self.mem_tbl_name
            ))


